// Test program for interval<bool>.
// Sylvain Pion, 2005.

#include <iostream>
#include <cassert>
#include "interval.h"

typedef std2::interval<bool>       Ib;

// checks that we can disallow arithmetic operators
// not being templates, these will always be compiled! :-(
// hence BOOST_STATIC_ASSERT(false) does not work
Ib operator+(Ib const&, Ib const&) { assert( false ); }
Ib operator+(bool const&, Ib const&) { assert( false ); }
Ib operator+(Ib const&, bool const&) { assert( false ); }

// Returns true if the conversion of i to bool throws.
bool throws(Ib const& i)
{
  try {
    (bool) i;
    return false;
  }
  catch (...) {
    return true;
  }
}


int main()
{
  Ib t(true), f(false), i(false, true), e(true, false);
  Ib ee, ii(Ib::indeterminate());

  // the next line would trigger an assert( false )
  // i + e + t + f;

  assert( is_indeterminate(i) );
  assert( is_indeterminate(ii) );
  assert( e.empty() );
  assert( ee.empty() );
  assert( t );
  assert( ! f );

  assert( equal( t, true) );
  assert( equal(!f, true) );
  assert( equal(!t, false) );
  assert( equal( f, false) );
  assert( equal(i, ii) );
  assert( equal(i, !ii) );
  assert( equal(e, ee) );
  assert( equal(e, !ee) );
  assert( ! equal( t, i) );
  assert( ! equal( f, i) );
  assert( ! equal( t, e) );
  assert( ! equal( f, e) );
  assert( ! equal( i, e) );

  assert( t == true );
  assert( f != true );
  assert( t != false );
  assert( f == false );
  assert( is_indeterminate(i == ii) );
  assert( is_indeterminate(i == !ii) );
  assert( (e == ee).empty() );
  assert( (e != ee).empty() );
  assert( is_indeterminate( t == i) );
  assert( is_indeterminate( f == i) );
  assert( ( t == e).empty() );
  assert( ( f == e).empty() );
  assert( ( i == e).empty() );

  assert( throws(i) );
  assert( throws(ii) );

  assert( t || true );
  assert( t || t );
  assert( t || false );
  assert( t || f );
  assert( t || i );
  assert( (t || e).empty() );

  assert( f || true );
  assert( f || t );
  assert( ! ( f || false ) );
  assert( ! ( f || f ) );
  assert( is_indeterminate(f || i) );
  assert( (f || e).empty() );

  assert( is_indeterminate(i || i) );
  assert( (i || e).empty() );

  assert( t && true );
  assert( t && t );
  assert( ! ( t && false ) );
  assert( ! ( t && f ) );
  assert( is_indeterminate( t && i ) );
  assert( (t && e).empty() );

  assert( ! ( f && true ) );
  assert( ! ( f && t ) );
  assert( ! ( f && false ) );
  assert( ! ( f && f ) );
  assert( ! ( f && i ) );
  assert( (f && e).empty() );

  assert( is_indeterminate(i && i) );
  assert( (i && e).empty() );

  return 0;
}
