
#include <iostream>
#include <cassert>
#include "interval.h"

typedef std2::interval<float>       If;
typedef std2::interval<double>      Id;
typedef std2::interval<long double> Il;

template < typename I >
void test()
{
  typedef typename I::value_type T;

  // Constructors
  I e, a(T(0)), b(T(2)), c(T(-1), T(1));

  assert( e.empty() );
  assert( a.inf() == T(0)  && a.sup() == T(0) );
  assert( b.inf() == T(2)  && b.sup() == T(2) );
  assert( c.inf() == T(-1) && c.sup() == T(1) );

  // I/O
  std::cout << "Check the output : " << "[ empty  ]  == " << e << std::endl;
  std::cout << "Check the output : " << "[  0 ; 0 ]  == " << a << std::endl;
  std::cout << "Check the output : " << "[ -1 ; 1 ]  == " << c << std::endl;

  // Accessory functions
  assert(  equal(I(0, 0), I(0, 0)) );
  assert( !equal(I(0, 0), I(1, 1)) );
  assert(  equal(I(0, 1), I(0, 1)) );
  assert( !equal(I(0, 1), I(0, 0)) );
  assert( !equal(I(0, 1), I(1, 1)) );

  assert(  is_singleton(a) );
  assert(  is_singleton(I(1, 1)) );
  assert( !is_singleton(I(0, 1)) );

  contains(a, T(0));
  contains(a, a);

  assert( width(a) == T(0) );
  assert( width(c) == T(2) );

  assert( intersect(a, a) == a );

  // Comparison operators
  assert( I(0,0) <  I(1,1)); assert( I(0,0) <  T(1));
  assert( I(0,0) <= I(1,1)); assert( I(0,0) <= T(1));
  a < b; a <= b; a > b; a >= b; a == b; a != b;
  a < T(1); a <= T(1); a > T(1); a >= T(1); a == T(1); a != T(1);
  T(1) < b; T(1) <= b; T(1) > b; T(1) >= b; T(1) == b; T(1) != b;

  // Arithmetic operators
  assert( equal(a, +a) );
  assert( equal(c, +c) );
  assert( equal(-a, I(-a.sup(), -a.inf())) );
  assert( equal(-c, I(-c.sup(), -c.inf())) );

  // TODO : check results
  a+b; a-b; a*b; a/b;
  a+T(1); a-T(1); a*T(1); a/T(1);
  T(1)+a; T(1)-a; T(1)*a; T(1)/b;
  a += a; a -= a; a *= a; a /= b;
  a += T(1); a -= T(1); a *= T(1); a /= T(1);
}

void test_conversions()
{
  // Check existence
  If f (1, 1);
  Id d (f);
  Il l (f), l1 (d);
  If f1 (d), f0 (l);
  Id d1 (l);

  // Check semantics
  If f2 = f/3;
  Id d2(f2);
  Il l2(f2);
  assert(! is_singleton(f2));
  assert(f2.inf() == d2.inf() && f2.sup() == d2.sup());
  assert(f2.inf() == l2.inf() && f2.sup() == l2.sup());

  Id d3 = d/3;
  If f3(d3);
  Il l3(d3);
  assert(! is_singleton(d3));
  assert(f3.inf() <= d3.inf() && f3.sup() >= d3.sup());
  assert(d3.inf() == l3.inf() && d3.sup() == l3.sup());

  Il l4 = l/3;
  If f4(l4);
  Id d4(l4);
  assert(! is_singleton(l4));
  assert(f4.inf() <= d4.inf() && f4.sup() >= d4.sup());
  assert(f4.inf() <= l4.inf() && f4.sup() >= l4.sup());
  assert(d4.inf() <= l4.inf() && d4.sup() >= l4.sup());

  std::cout << f2 << std::endl;
  std::cout << d3 << std::endl;
  std::cout << l4 << std::endl;
}

int main()
{
  std::cout.precision(20);

  test_conversions();

  // Built-in floating point types.
  test<If>();
  test<Id>();
  test<Il>();

  // What about built-in integral types ?
  // Should they be allowed ?  What about unsigned * ?  Which rounding ?
  // test<std2::interval<long long> >();
  // test<std2::interval<long> >();
  // test<std2::interval<int> >();
  // test<std2::interval<short> >();

  // std::complex<interval<T> > could make sense,
  // but std::complex does not allow this at the moment.

  // std2::interval<std::integer> would make sense, but is not there yet.
  // Same for std2::interval<std2::rational>.

  return 0;
}
